/**
 * Create new cross compiler object.
 */
function CrossCompiler()
{
    console.log(":: Cross compilation");
}

/**
 * Compile the source to flat description.
 */
CrossCompiler.prototype.compile = function(source)
{
    var parser = new CrossParser();
    var behaviour = parser.readBehaviour(source);
    var flatBehaviour = this.resolveDominancy(behaviour);
    var target = this.printBehaviourDescription(flatBehaviour);
    return target;
}

/**
 * Resolve the dominancy and construct flat behaviour description graph.
 */
CrossCompiler.prototype.resolveDominancy = function(behaviour)
{
    var flatBehaviour = {};

    var universes = behaviour["universes"];
    var rulebases = behaviour["rulebases"];

    flatBehaviour["universes"] = universes;
    flatBehaviour["rulebases"] = {}

    for (var rulebaseName in rulebases) {
        if (rulebases.hasOwnProperty(rulebaseName)) {
            var flatRules = this.getFlatRules(rulebases[rulebaseName], universes);
            flatBehaviour["rulebases"][rulebaseName] = {"rules": flatRules};
        }
    }

    return flatBehaviour;
}

/**
 * Get the flat rules by applying dominancy.
 */
CrossCompiler.prototype.getFlatRules = function(dominates, universes)
{
    if ("dominates" in dominates == false) {
        // NOTE: The dominates must contains the dominated rules!
        return dominates["rules"];
    }
    else {
        var dominatedRules = this.getFlatRules(dominates["dominates"], universes);
        var negatedPredicates = this.calcNegatedPredicates(dominates["rules"], universes);
        var extendedRules = this.extendsPredicates(dominatedRules, negatedPredicates);

        var flatRules = dominates["rules"];
        for (var i = 0; i < extendedRules.length; ++i) {
            flatRules.push(extendedRules[i]);
        }

        return flatRules;
    }
}

/**
 * Get rules from predicates and negate them.
 */
CrossCompiler.prototype.calcNegatedPredicates = function(rules, universes)
{
    var predicates = this.getPredicates(rules);
    var negatedPredicates = this.negatePredicates(predicates, universes);
    return negatedPredicates;
}

/**
 * Get the predicates part of the rules.
 */
CrossCompiler.prototype.getPredicates = function(rules)
{
    var predicates = [];
    for (var i = 0; i < rules.length; ++i) {
        predicates.push(rules[i]["predicates"]);
    }
    return predicates;
}

/**
 * Calculate the negation of the predicates.
 * This function transform back the expression to disjunctive normal form.
 *
 * There are conjunctions in rows and disjunction between rows.
 */
CrossCompiler.prototype.negatePredicates = function(predicates, universes)
{
    var negated = [];

    var negatedByElements = this.negatePredicatesByElements(predicates, universes);

    var count = this.countPredicateElements(predicates);
    var nPatterns = Math.pow(2, count);

    for (var n = 0; n < nPatterns; ++n) {
        var pattern = this.getBitPattern(count, n);
        var selectedPredicates = this.collectSelectedPredicates(negatedByElements, pattern);

        // if (this.hasEmptyRow(selectedPredicates) == false) {
        if (this.allRowHasOneElement(selectedPredicates) == true) {
            var conjunction = this.createConjunction(selectedPredicates);
            negated.push(conjunction);
        }
    }

    return negated;
}

/**
 * Negate the predicates by elements.
 * The meaning of the resulted expression does not remained the same.
 */
CrossCompiler.prototype.negatePredicatesByElements = function(predicates, universes)
{
    var negatedPredicates = [];
    for (var i = 0; i < predicates.length; ++i) {
        var row = {};
        for (var antecedent in predicates[i]) {
            var value = predicates[i][antecedent];
            var negatedValue = this.negateValue(antecedent, value, universes); 
            row[antecedent] = negatedValue;
        }
        negatedPredicates.push(row);
    }

    return negatedPredicates;
}

/**
 * Negate the predicate value on the given universe.
 */
CrossCompiler.prototype.negateValue = function(antecedent, negableSymbol, universes)
{
    if (antecedent in universes == false) {
        console.log("The antecedent is missing from the universe: '" + antecedent + "'");
        return;
    }

    var symbols = universes[antecedent]["symbols"];

    // TODO: Check the number of symbols in the universe!

    var minPosition = null;
    var maxPosition = null;
    var minSymbol = "";
    var maxSymbol = "";

    for (var symbol in symbols) {
        if (symbols.hasOwnProperty(symbol)) {
            var position = symbols[symbol]["position"];
            if (minSymbol == "") {
                minSymbol = symbol;
                maxSymbol = symbol;
                minPosition = position;
                maxPosition = position;
            }
            else {
                if (position < minPosition) {
                    minSymbol = symbol;
                    minPosition = position;
                } 
                else if (position > maxPosition) {
                    maxSymbol = symbol;
                    maxPosition = position;
                }
            }
        }
    }
    
    if (negableSymbol == minSymbol) {
        return maxSymbol;
    }
    else if (negableSymbol == maxSymbol) {
        return minSymbol;
    }
    else {
        // NOTE: Something wrong!
    }
}

/**
 * Get the bit pattern of the value.
 */
CrossCompiler.prototype.getBitPattern = function(length, value)
{
    // TODO: Should check the length parameter!
    var pattern = [];
    var v = value;
    for (var i = 0; i < length; ++i) {
        if (v % 2 == 0) {
            pattern.unshift(false);
        }
        else {
            pattern.unshift(true);
        }
        v = Math.floor(v / 2);
    }
    return pattern;
}

/**
 * Collect the selected predicates from predicates array.
 */
CrossCompiler.prototype.collectSelectedPredicates = function(predicates, pattern)
{
    // TODO: Check the size of the predicates and pattern array!
    var selectedPredicates = [];
    var index = 0;
    for (var row = 0; row < predicates.length; ++row) {
        var selected = {};
        for (var antecedent in predicates[row]) {
            if (predicates[row].hasOwnProperty(antecedent)) {
                if (pattern[index] == true) {
                    selected[antecedent] = predicates[row][antecedent];
                }
                ++index;
            }
        }
        selectedPredicates.push(selected);
    }
    return selectedPredicates;
}

/**
 * Count the elements in the predicate array.
 */
CrossCompiler.prototype.countPredicateElements = function(predicates)
{
    var count = 0;
    for (var row = 0; row < predicates.length; ++row) {
        count += Object.keys(predicates[row]).length;
    }
    return count;
}

/**
 * Check that all row has exactly one element.
 */
CrossCompiler.prototype.allRowHasOneElement = function(predicates)
{
    for (var row = 0; row < predicates.length; ++row) {
        if (Object.keys(predicates[row]).length != 1) {
            return false;
        }
    }

    return true;
}

/**
 * Check that there is an empty row or not.
 */
// NOTE: Not used in the current version!
CrossCompiler.prototype.hasEmptyRow = function(predicates)
{
    if (predicates.length == 0) {
        return true;
    }

    for (var row = 0; row < predicates.length; ++row) {
        if (Object.keys(predicates[row]).length == 0) {
            return true;
        }
    }
    return false;
}

/**
 * Create conjunction from predicate array.
 */
CrossCompiler.prototype.createConjunction = function(predicates)
{
    var conjunction = {};

    for (var row = 0; row < predicates.length; ++row) {
        for (var antecedent in predicates[row]) {
            if (predicates[row].hasOwnProperty(antecedent)) {
                if (antecedent in conjunction) {
                    // NOTE: Something wrong!
                }
                else {
                    conjunction[antecedent] = predicates[row][antecedent];
                }
            }
        }
    }

    return conjunction;
}

/**
 * Extends the dominated rules by the given predicates.
 */
CrossCompiler.prototype.extendsPredicates = function(rules, extensions)
{
    var extendedRules = [];

    for (var i = 0; i < rules.length; ++i) {
        var rule = rules[i];

        for (var j = 0; j < extensions.length; ++j) {
            var extension = extensions[j];

            var newRule = {};
            if ("consequent" in rule) {
                newRule["consequent"] = rule["consequent"];
            }
            else {
                newRule["variable"] = rule["variable"];
            }

            newRule["predicates"] = {};
            
            if ("predicates" in rule) {
                for (var antecedent in rule["predicates"]) {
                    if (rule["predicates"].hasOwnProperty(antecedent)) {
                        newRule["predicates"][antecedent] = rule["predicates"][antecedent];
                    }
                }
            }

            for (var antecedent in extension) {
                if (extension.hasOwnProperty(antecedent)) {
                    newRule["predicates"][antecedent] = extension[antecedent];
                }
            }

            extendedRules.push(newRule);
        }
    }

    return extendedRules;
}

/**
 * Print the content of the flat behaviour description.
 */
CrossCompiler.prototype.printBehaviourDescription = function(behaviour)
{
    var description = "";
    description += this.printUniverses(behaviour);
    description += this.printRulebases(behaviour);
    return description;
}

/**
 * Print the universes of the behaviour.
 */
CrossCompiler.prototype.printUniverses = function(behaviour)
{
    var universes = behaviour["universes"];

    var text = "";

    for (var universeName in universes) {
        if (universes.hasOwnProperty(universeName)) {
            text += this.printUniverse(universeName, universes[universeName]);
        }
    }

    return text;
}

/**
 * Print the content of the universe.
 */
CrossCompiler.prototype.printUniverse = function(name, universe)
{
    var text = "universe \"" + name + "\"\n";

    var symbols = universe["symbols"];
    for (var symbol in symbols) {
        if (symbols.hasOwnProperty(symbol)) {
            text += "    \"" + symbol + "\" " + symbols[symbol]["position"] + " " + symbols[symbol]["value"] + "\n";
        }
    }

    text += "end\n\n";

    return text;
}

/**
 * Print the rulebases of the behaviour.
 */
CrossCompiler.prototype.printRulebases = function(behaviour)
{
    var rulebases = behaviour["rulebases"];

    var text = "";

    for (var rulebaseName in rulebases) {
        if (rulebases.hasOwnProperty(rulebaseName)) {
            text += this.printRulebase(rulebaseName, rulebases[rulebaseName]);
        }
    }

    return text;
}

/**
 * Print the content of the rulebase.
 */
CrossCompiler.prototype.printRulebase = function(name, rulebase)
{
    var text = "rulebase \"" + name + "\"\n";

    var rules = rulebase["rules"];
    for (var i = 0; i < rules.length; ++i) {
        var rule = rules[i];
        text += this.printRule(rule);
    }

    text += "end\n\n";

    return text;
}

/**
 * Print the rule.
 */
CrossCompiler.prototype.printRule = function(rule)
{
    var text = "    rule\n";

    if ("consequent" in rule) {
        text += "        \"" + rule["consequent"] + "\"";
    }
    else {
        text += "        use \"" + rule["variable"] + "\"";
    }

    if ("predicates" in rule) {
        text += " when";
    }

    text += "\n";

    var predicates = rule["predicates"];
    var isFirst = true;
    for (var antecedent in predicates) {
        if (predicates.hasOwnProperty(antecedent)) {
            if (isFirst == true) {
                isFirst = false;
            }
            else {
                text += " and\n";
            }
            text += "            \"" + antecedent + "\" is \"" + predicates[antecedent] + "\"";
        }
    }

    text += "\n";
    text += "    end\n";

    return text;
}
