/**
 * Create new parser object.
 */
function CrossParser()
{
}

/**
 * Read behaviour description from the text.
 */
CrossParser.prototype.readBehaviour = function(text)
{
}

/**
 * Read the behaviour definition from the text.
 */
CrossParser.prototype.readBehaviour = function(text)
{
    var behaviour = {};

    behaviour["universes"] = {};
    behaviour["rulebases"] = {};

    this.tokenizer = new CrossTokenizer();

    this.tokenizer.setText(text);

    var token = {};
    do {
        token = this.tokenizer.getNextToken();
        if (token["type"] == "keyword") {
            if (token["value"] == "universe") {
                var universe = this.readUniverse();
                if (universe["name"] in behaviour["universes"]) {
                    var cursor = this.tokenizer.getPosition();
                    throw {
                        "message": "The '" + universe["name"] + "' universe already defined!",
                        "row": cursor["row"],
                        "column": cursor["column"]
                    };
                }
                behaviour["universes"][universe["name"]] = universe["value"];
            }
            else if (token["value"] == "rulebase") {
                var rulebase = this.readRulebase();
                if (rulebase["name"] in behaviour["rulebases"]) {
                    var cursor = this.tokenizer.getPosition();
                    throw {
                        "message": "The '" + rulebase["name"] + "' rulebase already defined!",
                        "row": cursor["row"],
                        "column": cursor["column"]
                    };
                }
                behaviour["rulebases"][rulebase["name"]] = rulebase["value"];
            }
            else if (token["value"] == "init") {
                var init = this.readInit();
                // TODO: Check the consistency of the initialization block!
                behaviour["init"] = init;
            }
        }
        else if (token["type"] != "empty") {
            var cursor = this.tokenizer.getPosition();
            throw {
                "message": "The 'rulebase' keyword is missing!",
                "row": cursor["row"],
                "column": cursor["column"]
            };
        }
    } while (token["type"] != "empty");

    return behaviour;
}

/**
 * Read the universe definition.
 */
CrossParser.prototype.readUniverse = function()
{
    var universe = {};

    var token = this.tokenizer.getNextToken();
    if (token["type"] == "literal") {

        // Get the name.
        var name = token["value"];

        // Get the universe.
        var universe = new Universe();

        token = this.tokenizer.getNextToken();

        // Skip the description keyword.
        if (token["type"] == "keyword") {
            token = this.tokenizer.getNextToken();
            token = this.tokenizer.getNextToken();
        }

        while (token["type"] == "literal") {

            var symbol = token["value"];
            var position = null;
            var value = null;

            token = this.tokenizer.getNextToken();
            if (token["type"] == "number") {
                position = parseFloat(token["value"]);
            }
            else {
                var cursor = this.tokenizer.getPosition();
                throw {
                    "message": "Invalid position in universe definition!",
                    "row": cursor["row"],
                    "column": cursor["column"]
                };
            }

            token = this.tokenizer.getNextToken();
            if (token["type"] == "keyword" && token["value"] == ",") {
                token = this.tokenizer.getNextToken();
            }

            if (token["type"] == "number") {
                value = parseFloat(token["value"]);
            }
            else {
                var cursor = this.tokenizer.getPosition();
                throw {
                    "message": "Invalid value in universe definition!",
                    "row": cursor["row"],
                    "column": cursor["column"]
                };
            }

            universe.addSymbol(symbol, position, value);

            token = this.tokenizer.getNextToken();
        }

        if (token["type"] == "keyword" && token["value"] == "end") {
            return {"name": name, "value": universe};
        }
        else {
            var cursor = this.tokenizer.getPosition();
            throw {
                "message": "The 'end' keyword is missing at the end of universe definition!",
                "row": cursor["row"],
                "column": cursor["column"]
            };
        }

    }
    else {
        var cursor = this.tokenizer.getPosition();
        throw {
            "message": "The name of the universe is missing!",
            "row": cursor["row"],
            "column": cursor["column"]
        };
    }
}

/**
 * Read the rulebase definition.
 */
CrossParser.prototype.readRulebase = function()
{
    var rulebase = {};

    var rulebaseName = "";
    var rulebaseValue = {};

    var token = this.tokenizer.getNextToken();
    if (token["type"] == "literal") {
        rulebaseName = token["value"]; 
        rulebaseValue = this.readDominates();
    }
    else {
        var cursor = this.tokenizer.getPosition();
        throw {
            "message": "The rulebase name is missing!",
            "row": cursor["row"],
            "column": cursor["column"]
        };
    }

    return {"name": rulebaseName, "value": rulebaseValue};
}

/**
 * Read dominated block.
 */
CrossParser.prototype.readDominates = function()
{
    var dominates = {};
    var rules = [];

    do {
        var token = this.tokenizer.getNextToken();

        if (token["type"] == "keyword") {
            if (token["value"] == "rule") {
                var rule = this.readRule();
                rules.push(rule);
            }
            else if (token["value"] == "dominates") {
                var inner = this.readDominates();
                dominates["dominates"] = inner;
            }
            else if (token["value"] == "description") {
                token = this.tokenizer.getNextToken();
                if (token["type"] != "literal") {
                    var cursor = this.tokenizer.getPosition();
                    throw {
                        "message": "Missing description text!",
                        "row": cursor["row"],
                        "column": cursor["column"]
                    };
                }
            }
            else if (token["value"] != "end") {
                var cursor = this.tokenizer.getPosition();
                throw {
                    "message": "Invalid keyword in rule definition!",
                    "row": cursor["row"],
                    "column": cursor["column"]
                };
            }
        }

    } while (token["type"] != "keyword" || token["value"] != "end");

    dominates["rules"] = rules;

    return dominates;
}

/**
 * Read the rule.
 */
CrossParser.prototype.readRule = function()
{
    var rule = {};

    var token = this.tokenizer.getNextToken();
    if (token["type"] == "literal" || token["type"] == "keyword") {

        if (token["type"] == "literal") {
            rule["consequent"] = token["value"];
        }
        else if (token["type"] == "keyword") {
            if (token["value"] == "use") {
                token = this.tokenizer.getNextToken();
                if (token["type"] == "literal") {
                    rule["variable"] = token["value"];
                }
                else {
                    // ERROR: Literal required!
                }
            }
            else {
                // ERROR: Invalid keyword! (Only 'use' accepted yet!)
            }
        }

        token = this.tokenizer.getNextToken();
        if (token["type"] == "keyword" && token["value"] == "when") {
            do {
                var predicateAntecedent = "";
                var predicateValue = "";

                token = this.tokenizer.getNextToken();
                if (token["type"] == "literal") {
                    predicateAntecedent = token["value"];
                }
                else {
                    var cursor = this.tokenizer.getPosition();
                    throw {
                        "message": "Literal required on antecedent side!",
                        "row": cursor["row"],
                        "column": cursor["column"]
                    };
                }

                token = this.tokenizer.getNextToken();
                if (token["type"] == "keyword" && token["value"] == "is") {
                }
                else {
                    var cursor = this.tokenizer.getPosition();
                    throw {
                        "message": "Missing 'is' keyword!",
                        "row": cursor["row"],
                        "column": cursor["column"]
                    };
                }

                token = this.tokenizer.getNextToken();
                if (token["type"] == "literal") {
                    predicateValue = token["value"];
                }
                else {
                    var cursor = this.tokenizer.getPosition();
                    throw {
                        "message": "The consequent value must be literal!",
                        "row": cursor["row"],
                        "column": cursor["column"]
                    };
                }

                if ("predicates" in rule == false) {
                    rule["predicates"] = {};
                }

                if (predicateAntecedent in rule["predicates"]) {
                    throw {"message": "The '" + predicateAntecedent + "' antecedent is duplicated in rule!"};
                }

                rule["predicates"][predicateAntecedent] = predicateValue;

                token = this.tokenizer.getNextToken();
                if (token["type"] != "keyword") {
                    var cursor = this.tokenizer.getPosition();
                    throw {
                        "message": "'and' or 'end' keyword is required instead of literal!",
                        "row": cursor["row"],
                        "column": cursor["column"]
                    };
                }

            } while (token["value"] != "end");
        }
        else {
            // NOTE: Rule without predicates.
            if (token["type"] != "keyword" || token["value"] != "end") {
                var cursor = this.tokenizer.getPosition();
                throw {
                    "message": "The default rule also should be ended with 'end' keyword!",
                    "row": cursor["row"],
                    "column": cursor["column"]
                };
            }
        }
    }
    else {
        var cursor = this.tokenizer.getPosition();
        throw {
            "message": "Missing consequent value from rule!",
            "row": cursor["row"],
            "column": cursor["column"]
        };
    }

    return rule;
}

/**
 * Read the initialization block.
 */
CrossParser.prototype.readInit = function()
{
    var init = {};

    do {
        var token = this.tokenizer.getNextToken();
        if (token["type"] == "text") {
            var name = token["value"];
            var assignment = {};
            token = this.tokenizer.getNextToken();
            if (token["type"] == "text") {
                assignment[name] = {"symbol": token["value"]};
            }
            else if (token["type"] == "number") {
                // QUEST: Where should parse the value?
                assignment[name] = {"value": token["value"]};
            }
            else {
                var cursor = this.tokenizer.getPosition();
                throw {
                    "message": "The initialization value should be a symbol or a number!",
                    "row": cursor["row"],
                    "column": cursor["column"]
                };
            }
        }
    } while (token["type"] != "keyword" || token["value"] != "end");

    return init;
}

