/**
 * Create new map.
 */
function Map()
{
    this.width = 0;
    this.height = 0;
    this.walls = [];
}

/**
 * Get resizer of a wall.
 *
 * Returns with null when there is no resizer under the mouse cursor.
 */
Map.prototype.getResizer = function()
{
    for (var i = 0; i < this.walls.length; ++i) {

        var wall = this.walls[i];

        // The maximal distance from the selected point.
        var epsilon = 6;

        var x = mousePosition["x"];
        var y = mousePosition["y"];

        var x1 = wall["x1"];
        var y1 = wall["y1"];
        var x2 = wall["x2"];
        var y2 = wall["y2"];

        if (x > x1 - epsilon && x < x1 + epsilon && y > y1 - epsilon && y < y1 + epsilon) {
            return new Resizer(this.walls[i], "11");
        }

        if (x > x2 - epsilon && x < x2 + epsilon && y > y1 - epsilon && y < y1 + epsilon) {
            return new Resizer(this.walls[i], "21");
        }

        if (x > x1 - epsilon && x < x1 + epsilon && y > y2 - epsilon && y < y2 + epsilon) {
            return new Resizer(this.walls[i], "12");
        }

        if (x > x2 - epsilon && x < x2 + epsilon && y > y2 - epsilon && y < y2 + epsilon) {
            return new Resizer(this.walls[i], "22");
        }
    }
    
    return null;
}

/**
 * Render the map.
 */
Map.prototype.render = function(context)
{
    for (var i = 0; i < this.walls.length; ++i) {
        var wall = this.walls[i];

        var wallWidth = wall["x2"] - wall["x1"];
        var wallHeight = wall["y2"] - wall["y1"];

        context.beginPath();
        context.rect(wall["x1"], wall["y1"], wallWidth, wallHeight);
        context.fillStyle = '#EEEEEE';
        context.fill();
        context.lineWidth = 1;
        context.strokeStyle = 'black';
        context.stroke();

        this.renderResizers(wall, context);
    }
}

/**
 * Draw the resizers if focused.
 */
Map.prototype.renderResizers = function(wall, context)
{
    // The maximal distance from the selected point.
    var epsilon = 6;

    var x = mousePosition["x"];
    var y = mousePosition["y"];

    var x1 = wall["x1"];
    var y1 = wall["y1"];
    var x2 = wall["x2"];
    var y2 = wall["y2"];

    var resizer = null;

    if (x > x1 - epsilon && x < x1 + epsilon && y > y1 - epsilon && y < y1 + epsilon) {
        resizer = {"x": x1, "y": y1};
    }

    if (x > x2 - epsilon && x < x2 + epsilon && y > y1 - epsilon && y < y1 + epsilon) {
        resizer = {"x": x2, "y": y1};
    }

    if (x > x1 - epsilon && x < x1 + epsilon && y > y2 - epsilon && y < y2 + epsilon) {
        resizer = {"x": x1, "y": y2};
    }

    if (x > x2 - epsilon && x < x2 + epsilon && y > y2 - epsilon && y < y2 + epsilon) {
        resizer = {"x": x2, "y": y2};
    }

    if (resizer != null) {
        var size = epsilon * 2;
        context.beginPath();
        context.rect(resizer["x"] - epsilon, resizer["y"] - epsilon, size, size);
        context.fillStyle = '#888888';
        context.fill();
        context.lineWidth = 1;
        context.strokeStyle = 'black';
        context.stroke();
    }
}

/**
 * Set the size of the map.
 */
Map.prototype.setSize = function(width, height)
{
    this.width = width;
    this.height = height;
}

/**
 * Add new rectangular wall at the given position.
 */
Map.prototype.addWall = function(wall)
{
    this.walls.push(wall);
}

/**
 * Check that the given position is accessible with the radius.
 */
Map.prototype.isAccessible = function(x, y, radius)
{
    if (x < 0 || x > this.width || y < 0 || y > this.height) {
        return false;
    }

    for (var i = 0; i < this.walls.length; ++i) {
        var wall = this.walls[i];
        var x1 = wall["x1"] - radius;
        var y1 = wall["y1"] - radius;
        var x2 = wall["x2"] + radius;
        var y2 = wall["y2"] + radius;
        if (x >= x1 && x <= x2 && y > y1 && y < y2) {
            return false;
        }
    }
    return true;
}

/**
 * Calculate the closest point 
 */
Map.prototype.calcClosestWallPoint = function(x, y)
{
    var wallDistance = 1000;
    var closestPoint = null;

    for (var i = 0; i < this.walls.length; ++i) {
        var wall = this.walls[i];
        var point = this.findClosestWallPoint(x, y, wall);
        var distance = this.calcPointDistance(x, y, point);
        if (distance < wallDistance) {
            wallDistance = distance;
            closestPoint = point;
        }
    }

    return closestPoint;
}

/**
 * Calculate the distance from the wall.
 */
Map.prototype.calcPointDistance = function(x, y, point)
{
    var dx = x - point["x"];
    var dy = y - point["y"];
    var distance = Math.sqrt(dx * dx + dy * dy);
    return distance;
}

/**
 * Find the closest point on the wall.
 */
Map.prototype.findClosestWallPoint = function(x, y, wall)
{
    var x1 = wall["x1"];
    var y1 = wall["y1"];
    var x2 = wall["x2"];
    var y2 = wall["y2"];

    if (x > x1 && x < x2) {
        if (y < y1) {
            return {"x": x, "y": y1};
        }
        else if (y > y2) {
            return {"x": x, "y": y2};
        }
    }
    else {
        if (y > y1 && y < y2) {
            if (x < x1) {
                return {"x": x1, "y": y};
            }
            else if (x > x2) {
                return {"x": x2, "y": y};
            }
        }
        else {
            if (y < y1) {
                if (x < x1) {
                    return {"x": x1, "y": y1};
                }
                else if (x > x2) {
                    return {"x": x2, "y": y1};
                }
            }
            else if (y > y2) {
                if (x < x1) {
                    return {"x": x1, "y": y2};
                }
                else if (x > x2) {
                    return {"x": x2, "y": y2};
                }
            }
        }
    }

    return {"x": (x1 + x2) / 2, "y": (y1 + y2) / 2};
}

