describe("Universe", function() {

    var universe;

    beforeEach(function() {
        universe = new Universe();
    });

    describe("Empty", function() {

        it("Should be zero elements of empty universe", function() {
            expect(universe.countSymbols()).toEqual(0);
        });

        it("Should throw exception when try to calculate value", function() {
            var testFn = function() {
                universe.calcValue(0.0);
            }
            expect(testFn).toThrow({"message": "Unable to calc on empty universe!"});
        });

    });

    describe("Insertion", function() {

        it("Insert symbol into the universe", function() {
            universe.addSymbol("low", 0, 0);
            expect(universe.countSymbols()).toEqual(1);
        });

        it("Check symbol structure", function() {
            universe.addSymbol("low", 0, 0);
            expect("low" in universe.symbols).toEqual(true);
            expect(universe.symbols["low"]).toEqual({"position": 0.0, "value": 0.0});
        });

        it("Duplicated symbol", function() {
            universe.addSymbol("low", 0, 0);
            var testFn = function() {
                universe.addSymbol("low", 0, 0);
            }
            expect(testFn).toThrow({"message": "The 'low' symbol has been already defined in universe!"});
        });

        it("Throw exception on duplicated position", function() {
            var testFn = function() {
                universe.addSymbol("low", 0.2, 0);
                universe.addSymbol("high", 0.2, 1);
            }
            expect(testFn).toThrow({"message": "Duplicated position in the universe!"});
        });

        it("Insert multiple symbols", function() {
            universe.addSymbol("low", 0.0, 0.0);
            universe.addSymbol("medium", 0.5, 0.5);
            universe.addSymbol("high", 1.0, 1.0);
            expect(universe.countSymbols()).toEqual(3);
        });

        it("Check monotonity", function() {
            var testFn = function() {
                universe.addSymbol("low", 0.0, 0.5);
                universe.addSymbol("high", 1.0, 0.2);
            }
            expect(testFn).toThrow({"message": "The universe function must be monoton!"});
        });

    });

    describe("Low and high positions", function() {

        it("Low symbol", function() {
            universe.addSymbol("low", 0, 0);
            universe.addSymbol("high", 1, 1);
            expect(universe.getLowPosition()).toEqual(0);
        });

        it("Low symbol (reverse order)", function() {
            universe.addSymbol("high", 1, 1);
            universe.addSymbol("low", 0, 0);
            expect(universe.getLowPosition()).toEqual(0);
        });

        it("High symbol", function() {
            universe.addSymbol("low", 0, 0);
            universe.addSymbol("high", 1, 1);
            expect(universe.getHighPosition()).toEqual(1);
        });

        it("High symbol (reverse order)", function() {
            universe.addSymbol("high", 1, 1);
            universe.addSymbol("low", 0, 0);
            expect(universe.getHighPosition()).toEqual(1);
        });

    });

    describe("Missing symbol definitions for calculation", function() {

        it("Empty universe - find closest on left", function() {
            var testFn = function() {
                universe.findClosestSymbolOnLeft(0.5);
            }
            expect(testFn).toThrow({"message": "Cannot find symbol on empty universe!"});
        });

        it("Empty universe - find closest on right", function() {
            var testFn = function() {
                universe.findClosestSymbolOnRight(0.5);
            }
            expect(testFn).toThrow({"message": "Cannot find symbol on empty universe!"});
        });

        it("Single value at middle - find closest on left", function() {
            universe.addSymbol("middle", 0.5, 0.4);
            var testFn = function() {
                universe.findClosestSymbolOnLeft(0.1);
            }
            expect(testFn).toThrow({"message": "At least two symbols should be defined on the universe!"});
        });

        it("Single value at middle - find closest on right", function() {
            universe.addSymbol("middle", 0.5, 0.4);
            var testFn = function() {
                universe.findClosestSymbolOnRight(0.1);
            }
            expect(testFn).toThrow({"message": "At least two symbols should be defined on the universe!"});
        });

    });

    describe("Out of universe range", function() {

        beforeEach(function() {
            universe.addSymbol("low", 0.0, 0.0);
            universe.addSymbol("high", 1.0, 1.0);
        });

        it("Too small value - on left", function() {
            var testFn = function() {
                universe.findClosestSymbolOnLeft(-0.1);
            }
            expect(testFn).toThrow({"message": "The value is below the universe bounds! (-0.1 < 0)"});
        });

        it("Too small value - on right", function() {
            var testFn = function() {
                universe.findClosestSymbolOnRight(-0.1);
            }
            expect(testFn).toThrow({"message": "The value is below the universe bounds! (-0.1 < 0)"});
        });

        it("Too large value - on left", function() {
            var testFn = function() {
                universe.findClosestSymbolOnLeft(1.1);
            }
            expect(testFn).toThrow({"message": "The value is above the universe bounds! (1.1 > 1)"});
        });

        it("Too large value - on right", function() {
            var testFn = function() {
                universe.findClosestSymbolOnRight(1.1);
            }
            expect(testFn).toThrow({"message": "The value is above the universe bounds! (1.1 > 1)"});
        });

    });

    describe("Find closest symbols from five symbols", function() {

        it("Ordered symbols", function() {

            universe.addSymbol("a", 0.1, 1);
            universe.addSymbol("b", 0.3, 2);
            universe.addSymbol("c", 0.6, 3);
            universe.addSymbol("d", 0.8, 4);
            universe.addSymbol("e", 1.3, 5);

            expect(universe.findClosestSymbolOnLeft(0.2)).toEqual("a");
            expect(universe.findClosestSymbolOnRight(0.2)).toEqual("b");

            expect(universe.findClosestSymbolOnLeft(0.5)).toEqual("b");
            expect(universe.findClosestSymbolOnRight(0.5)).toEqual("c");

            expect(universe.findClosestSymbolOnLeft(0.7)).toEqual("c");
            expect(universe.findClosestSymbolOnRight(0.7)).toEqual("d");

            expect(universe.findClosestSymbolOnLeft(1)).toEqual("d");
            expect(universe.findClosestSymbolOnRight(1)).toEqual("e");

        });

    });

    describe("Calculation with low and high symbols", function() {

        beforeEach(function() {
            universe.addSymbol("low", 0, 0);
            universe.addSymbol("high", 1, 1);
        });

        it("Low value", function() {
            expect(universe.calcValue(0.0)).toEqual(0.0);
        });

        it("High value", function() {
            expect(universe.calcValue(1.0)).toEqual(1.0);
        });

        it("Inner values", function() {
            for (var i = 0; i < 100; ++i) {
                var observation = i / 100;
                expect(universe.calcValue(observation)).toEqual(observation);
            }
        });

    });

    describe("Calculation with three values", function() {

        beforeEach(function() {
            universe.addSymbol("low", 0, 0);
            universe.addSymbol("middle", 0.5, 0.2);
            universe.addSymbol("high", 1, 1);
        });

        it("Inner values", function() {
            for (var i = 0; i < 100; ++i) {
                var observation = i / 200;
                var value = i / 500;
                expect(universe.calcValue(observation)).toBeCloseTo(value, 8);
            }
            for (var i = 0; i < 100; ++i) {
                var observation = 0.5 + i / 200;
                var value = 0.2 + i / 125;
                expect(universe.calcValue(observation)).toBeCloseTo(value, 8);
            }
        });

    });

    describe("Calculation with three values (shifted)", function() {

        beforeEach(function() {
            universe.addSymbol("low", 10, 0);
            universe.addSymbol("middle", 10.5, 0.2);
            universe.addSymbol("high", 11, 1);
        });

        it("Inner values", function() {
            for (var i = 0; i < 100; ++i) {
                var observation = i / 200 + 10;
                var value = i / 500;
                expect(universe.calcValue(observation)).toBeCloseTo(value, 8);
            }
            for (var i = 0; i < 100; ++i) {
                var observation = 0.5 + i / 200 + 10;
                var value = 0.2 + i / 125;
                expect(universe.calcValue(observation)).toBeCloseTo(value, 8);
            }
        });

    });

    describe("Calculation with three values (scaled)", function() {

        beforeEach(function() {
            universe.addSymbol("low", 0, 0);
            universe.addSymbol("middle", 0.5, 1);
            universe.addSymbol("high", 1, 5);
        });

        it("Inner values", function() {
            for (var i = 0; i < 100; ++i) {
                var observation = i / 200;
                var value = (i / 500) * 5;
                expect(universe.calcValue(observation)).toBeCloseTo(value, 8);
            }
            for (var i = 0; i < 100; ++i) {
                var observation = 0.5 + i / 200;
                var value = (0.2 + i / 125) * 5;
                expect(universe.calcValue(observation)).toBeCloseTo(value, 8);
            }
        });

    });

    describe("Rotation universe", function() {

        beforeEach(function() {
            universe.addSymbol("left", -180, -180);
            universe.addSymbol("no turn", 0, 0);
            universe.addSymbol("right", 180, 180);
        });

        it("Left rule distance", function() {
            expect(universe.calcDistance(-180, -180)).toBeCloseTo(0.0, 8);
            expect(universe.calcDistance(-180, 0)).toBeCloseTo(0.5, 8);
            expect(universe.calcDistance(-180, 180)).toBeCloseTo(1.0, 8);
        });

        it("No turn rule distance", function() {
            expect(universe.calcDistance(0, -180)).toBeCloseTo(0.5, 8);
            expect(universe.calcDistance(0, 0)).toBeCloseTo(0.0, 8);
            expect(universe.calcDistance(0, 180)).toBeCloseTo(0.5, 8);
        });

        it("Right rule distance", function() {
            expect(universe.calcDistance(180, -180)).toBeCloseTo(1.0, 8);
            expect(universe.calcDistance(180, 0)).toBeCloseTo(0.5, 8);
            expect(universe.calcDistance(180, 180)).toBeCloseTo(0.0, 8);
        });
        
    });

});
