/**
 * Create new world object.
 */
function World()
{
    this.initialize();
}

/**
 * Initialize the world.
 */
World.prototype.initialize = function()
{
    this.canvas = document.getElementById("world-canvas");
    this.context = this.canvas.getContext('2d');

    this.canvas.addEventListener("mousedown", this.mouseClick, false);

    this.map = new Map();
    this.destination = new Destination();
    this.actor = new Actor();
}

/**
 * Get mouse position from the mouse event.
 */
World.prototype.getMousePosition = function(event)
{
    var mouseX;
    var mouseY;

    if (event.pageX != undefined && event.pageY != undefined) {
        mouseX = event.pageX;
        mouseY = event.pageY;
    }
    else {
        mouseX = event.clientX + document.body.scrollLeft + document.documentElement.scrollLeft;
        mouseY = event.clientY + document.body.scrollTop + document.documentElement.scrollTop;
    }

    console.log(this.canvas);

    mouseX -= this.canvas.offsetLeft;
    mouseY -= this.canvas.offsetTop;

    return {"x": mouseX, "y": mouseY};
}

/**
 * Set the position of the destination.
 */
World.prototype.setDestinationPosition = function(x, y)
{
    if (this.map.isAccessible(x, y) == true) {
        this.destination.setPosition(x, y);
    }
}

/**
 * Calculate the view angle of the actor from destination.
 */
World.prototype.calcViewAngle = function()
{
    return this.calcAngle(this.actor.x, this.actor.y, this.actor.direction, this.destination.x, this.destination.y);
}

/**
 * Calculate the distance and angle from the closest wall.
 */
World.prototype.calcWallAngle = function()
{
    var closestPoint = this.map.calcClosestWallPoint(this.actor.x, this.actor.y);
    return this.calcAngle(this.actor.x, this.actor.y, this.actor.direction, closestPoint["x"], closestPoint["y"]);
}

/**
 * Calculate angle between an origin and an arbitrary point.
 */
World.prototype.calcAngle = function(originX, originY, direction, pointX, pointY)
{
    var dx = pointX - originX;
    var dy = pointY - originY;

    var distance = Math.sqrt(dx * dx + dy * dy);

    if (distance < 0.0001) {
        return 0.0;
    }

    var angle;
    if (Math.abs(dx) > Math.abs(dy)) {
        if (dx > 0.0) {
            angle = Math.atan(dy / dx);
        }
        else {
            angle = Math.PI + Math.atan(dy / dx);
            if (angle > Math.PI) {
                angle -= Math.PI * 2;
            }
        }
    }
    else {
        if (dy > 0.0) {
            angle = Math.PI / 2 - Math.atan(dx / dy);
        }
        else {
            angle = -(Math.atan(dx / dy) + Math.PI / 2);
        }
    }

    angle -= direction;

    while (angle < -Math.PI) {
        angle += Math.PI * 2;
    }

    while (angle > Math.PI) {
        angle -= Math.PI * 2;
    }

    return {"distance": distance, "angle": angle};
}

/**
 * Calculate the range ~ the distance from the wall.
 */
World.prototype.calcRange = function()
{
    var stepSize = 1.0;

    var x = this.actor.x;
    var y = this.actor.y;

    var vx = Math.cos(this.actor.direction) * stepSize;
    var vy = Math.sin(this.actor.direction) * stepSize;

    var range = 0.0;

    while (this.map.isAccessible(x, y, 0.0)) {
        range += stepSize;
        x += vx;
        y += vy;
    }

    return range;
}

/**
 * Update the world.
 */
World.prototype.update = function()
{
    // NOTE: Required for refreshing position when there is selected object!
    this.mouseMove();

    this.updateActorPosition();
    this.render();
}

/**
 * Calculate the new position of the actor.
 */
World.prototype.updateActorPosition = function()
{
    this.moveActor();
    this.rotateActor();
}

/**
 * Step the actor forward at given speed.
 *
 * NOTE: The speed can be negative.
 */
World.prototype.moveActor = function()
{
    // The required vector of moving.
    var vx = Math.cos(this.actor.direction) * this.actor.speed;
    var vy = Math.sin(this.actor.direction) * this.actor.speed;
    
    // Move and apply collision.
    var x = this.actor.x;
    var y = this.actor.y;

    if (this.map.isAccessible(x + vx, y, this.actor.radius)) {
        x += vx;
    }
    
    if (this.map.isAccessible(x, y + vy, this.actor.radius)) {
        y += vy;
    }

    // Set the accessible position.
    this.actor.setPosition(x, y);
}

/**
 * Rotate the actor with given angle.
 */
World.prototype.rotateActor = function()
{
    var direction = this.actor.direction + this.actor.rotation;
    while (direction < -Math.PI) {
        direction += Math.PI * 2;
    }
    while (direction > Math.PI) {
        direction -= Math.PI * 2;
    }
    this.actor.setDirection(direction);
}

/**
 * Render the world.
 */
World.prototype.render = function()
{
    this.context.clearRect(0, 0, this.canvas.width, this.canvas.height);

    this.map.render(this.context);
    this.destination.render(this.context);
    this.actor.render(this.context);
}

/**
 * Handle mouse down event.
 */
World.prototype.mouseDown = function()
{
    if (this.destination.isFocused() == true) {
        this.focused = this.destination;
    }
    else if (this.actor.isFocused() == true) {
        this.focused = this.actor;
    }
    else {
        this.focused = this.map.getResizer();
    }
}

/**
 * Handle mouse up event.
 */
World.prototype.mouseUp = function()
{
    this.focused = null;
}

/**
 * Handle mouse move event.
 */
World.prototype.mouseMove = function()
{
    if (this.focused != null) {
        this.focused.setPosition(mousePosition["x"], mousePosition["y"]);
    }
}

